import os
import re
import json
import argparse
import subprocess
import webbrowser
import platform
from pathlib import Path
from colorama import Fore, init

from plumai.app import app

# Initialize colorama
init(autoreset=True)

# Determine the cache directory based on the OS
def get_cache_directory() -> Path:
    if platform.system() == "Windows":
        cache_dir = Path(os.getenv("APPDATA", Path.home() / "AppData" / "Roaming")) / "plumai"
    elif platform.system() == "Darwin":  # macOS
        cache_dir = Path.home() / ".plumai"
    else:  # Assume Linux
        cache_dir = Path.home() / ".plumai"
    cache_dir.mkdir(parents=True, exist_ok=True)
    return cache_dir

CACHE_FILE = get_cache_directory() / "model_cache.json"


class PlumAIDeploymentError(Exception):
    pass


def print_plumai(message: str, color: str = Fore.WHITE) -> None:
    """Prints a message with a [plumai] prefix and optional color."""
    print(color + f"[plumai] {message}")


def _get_cached_model_endpoint(model_name: str) -> str:
    """Retrieve cached model endpoint from the cache file."""
    if CACHE_FILE.exists():
        with open(CACHE_FILE, "r") as f:
            cache = json.load(f)
            return cache.get(model_name)
    return None

def _cache_model_endpoint(model_name: str, endpoint: str) -> None:
    """Cache the model endpoint."""
    cache = {}
    if CACHE_FILE.exists():
        with open(CACHE_FILE, "r") as f:
            cache = json.load(f)
    cache[model_name] = endpoint
    with open(CACHE_FILE, "w") as f:
        json.dump(cache, f)


def _deploy(model_name: str, gpu_type: str, force: bool) -> str:
    cached_endpoint = _get_cached_model_endpoint(model_name)

    if not force and cached_endpoint:
        print_plumai(f"Model {model_name} is already deployed. Using cached endpoint.", Fore.GREEN)
        return cached_endpoint

    root_dir = Path(__file__).parent
    models_dir = (root_dir / "models").resolve()

    if model_name == "flux.1-dev":
        model_file = str(models_dir / "flux1_dev.py")
    elif model_name == "flux.1-schnell":
        model_file = str(models_dir / "flux1_schnell.py")
    else:
        raise ValueError(f"{model_name} not found")

    command = f"GPU_TYPE={gpu_type} modal deploy {model_file}"
    try:
        print_plumai(f"Deploying model {model_name} gpu: {gpu_type}.", Fore.BLUE)
        print_plumai(f"This may take a few minutes on the first deployment...", Fore.BLUE)
        result = subprocess.run(command, shell=True, check=True, capture_output=True, text=True)
        output = result.stdout

        link_pattern = r'https://[^\s"]+'
        match = re.search(link_pattern, output)
        if match:
            model_endpoint = match.group(0)
        else:
            raise PlumAIDeploymentError("Model endpoint not found.")

        _cache_model_endpoint(model_name, model_endpoint)
        print_plumai(f"Model {model_name} deployed successfully.", Fore.GREEN)
        print_plumai(f"Model endpoint: {model_endpoint}", Fore.GREEN)
        return model_endpoint
    except subprocess.CalledProcessError as e:
        print_plumai(f"Failed to deploy model {model_name}. Return code: {e.returncode}", Fore.RED)
        return
    except PlumAIDeploymentError as e:
        print_plumai(f"Error: {str(e)}", Fore.RED)
        return


def run(args):
    model_endpoint = _deploy(args.model_name, args.gpu_type, args.force)
    if model_endpoint:
        os.environ["MODEL_NAME"] = args.model_name
        os.environ["MODEL_ENDPOINT"] = model_endpoint
        port = 6868
        webbrowser.open(f"http://localhost:{port}")
        app.run(host="0.0.0.0", port=port)


def main():
    parser = argparse.ArgumentParser(prog="plumai")
    subparsers = parser.add_subparsers(dest="command")

    # Define `run` command
    run_parser = subparsers.add_parser("run", help="Deploy and run a model")
    run_parser.add_argument("model_name", type=str,
                            choices=["flux.1-dev", "flux.1-schnell"],
                            help="Model name. Available options: flux.1-dev (default), flux.1-schnell")
    run_parser.add_argument("gpu_type", type=str, nargs="?", default="A10G",
                            choices=["A10G", "A100"],
                            help="GPU type. Available options: A10G (default), A100")
    run_parser.add_argument("-f", "--force", action="store_true",
                            help="Force redeployment of the model, ignoring the cache")

    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
    else:
        run(args)


if __name__ == "__main__":
    main()


import os
import io

from modal import (
    App,
    Image,
    build,
    enter,
    gpu,
    method,
    web_endpoint
)


flux_image = (
    Image.from_registry(
        "nvidia/cuda:12.3.1-base-ubuntu22.04", add_python="3.10"
    )
    .apt_install(
        "git", "libglib2.0-0", "libsm6", "libxrender1", "libxext6", "ffmpeg", "libgl1", "aria2"
    )
    .run_commands("git clone -b totoro3 https://github.com/camenduru/ComfyUI /TotoroUI")
    .pip_install(
        "torchsde",
        "einops",
        "diffusers",
        "accelerate",
        "xformers",
        "torchvision",
        "transformers"
    )
)


app = App("flux.1-dev")

with flux_image.imports():
    import random
    import torch
    import numpy as np
    from PIL import Image

gpu_type = os.getenv("GPU_TYPE", "A10G")
@app.cls(gpu=gpu_type, container_idle_timeout=240, image=flux_image)
class Model:
    @build()
    def build(self):
        import os
        import subprocess

        os.chdir("/TotoroUI")

        subprocess.run("aria2c --console-log-level=error -c -x 16 -s 16 -k 1M https://huggingface.co/camenduru/FLUX.1-dev/resolve/main/flux1-dev-fp8.safetensors -d /TotoroUI/models/unet -o flux1-dev-fp8.safetensors", shell=True,)
        subprocess.run("aria2c --console-log-level=error -c -x 16 -s 16 -k 1M https://huggingface.co/camenduru/FLUX.1-dev/resolve/main/ae.sft -d /TotoroUI/models/vae -o ae.sft", shell=True,)
        subprocess.run("aria2c --console-log-level=error -c -x 16 -s 16 -k 1M https://huggingface.co/camenduru/FLUX.1-dev/resolve/main/clip_l.safetensors -d /TotoroUI/models/clip -o clip_l.safetensors", shell=True,)
        subprocess.run("aria2c --console-log-level=error -c -x 16 -s 16 -k 1M https://huggingface.co/camenduru/FLUX.1-dev/resolve/main/t5xxl_fp8_e4m3fn.safetensors -d /TotoroUI/models/clip -o t5xxl_fp8_e4m3fn.safetensors", shell=True,)

    @enter()
    def enter(self):
        import os
        import sys

        project_path = "/TotoroUI"
        if project_path not in sys.path:
            sys.path.append(project_path)

        os.chdir(project_path)

        import torch
        from nodes import NODE_CLASS_MAPPINGS
        from totoro_extras import nodes_custom_sampler

        self.DualCLIPLoader = NODE_CLASS_MAPPINGS["DualCLIPLoader"]()
        self.UNETLoader = NODE_CLASS_MAPPINGS["UNETLoader"]()
        self.RandomNoise = nodes_custom_sampler.NODE_CLASS_MAPPINGS["RandomNoise"]()
        self.BasicGuider = nodes_custom_sampler.NODE_CLASS_MAPPINGS["BasicGuider"]()
        self.KSamplerSelect = nodes_custom_sampler.NODE_CLASS_MAPPINGS["KSamplerSelect"]()
        self.BasicScheduler = nodes_custom_sampler.NODE_CLASS_MAPPINGS["BasicScheduler"]()
        self.SamplerCustomAdvanced = nodes_custom_sampler.NODE_CLASS_MAPPINGS["SamplerCustomAdvanced"]()
        self.VAELoader = NODE_CLASS_MAPPINGS["VAELoader"]()
        self.VAEDecode = NODE_CLASS_MAPPINGS["VAEDecode"]()
        self.EmptyLatentImage = NODE_CLASS_MAPPINGS["EmptyLatentImage"]()
        with torch.inference_mode():
            self.clip = self.DualCLIPLoader.load_clip("t5xxl_fp8_e4m3fn.safetensors", "clip_l.safetensors", "flux")[0]
            self.unet = self.UNETLoader.load_unet("flux1-dev-fp8.safetensors", "fp8_e4m3fn")[0]
            self.vae = self.VAELoader.load_vae("ae.sft")[0]

    def _inference(self, prompt, width, height, seed, guidance_scale, n_steps):
        import sys

        project_path = "/TotoroUI"
        if project_path not in sys.path:
            sys.path.append(project_path)

        from totoro import model_management

        with torch.inference_mode():
            sampler_name = "euler"
            scheduler = "simple"

            if seed < 0:
                seed = random.randint(0, 18446744073709551615)

            cond, pooled = self.clip.encode_from_tokens(self.clip.tokenize(prompt), return_pooled=True)
            cond = [[cond, {"pooled_output": pooled}]]
            noise = self.RandomNoise.get_noise(seed)[0]
            guider = self.BasicGuider.get_guider(self.unet, cond)[0]
            sampler = self.KSamplerSelect.get_sampler(sampler_name)[0]
            sigmas = self.BasicScheduler.get_sigmas(self.unet, scheduler, n_steps, 1.0)[0]
            latent_image = self.EmptyLatentImage.generate(closestNumber(width, 16), closestNumber(height, 16))[0]
            sample, sample_denoised = self.SamplerCustomAdvanced.sample(noise, guider, sampler, sigmas, latent_image)
            model_management.soft_empty_cache()
            decoded = self.VAEDecode.decode(self.vae, sample)[0].detach()
            image = Image.fromarray(np.array(decoded*255, dtype=np.uint8)[0])

            byte_stream = io.BytesIO()
            image.save(byte_stream, format="JPEG")
            return byte_stream

    @method()
    def inference(self, prompt, width, height, seed, guidance_scale, n_steps):
        return self._inference(prompt, width, height, seed, guidance_scale, n_steps).getvalue()

    @web_endpoint(docs=True)
    def web_inference(
        self,
        prompt: str,
        width: int = 1024,
        height: int = 1024,
        seed: int = -1,
        guidance_scale: float = 3.5,
        n_steps: int = 20
    ):
        from fastapi import Response
        return Response(
            content=self._inference(
                prompt,
                width,
                height,
                seed,
                guidance_scale,
                n_steps
            ).getvalue(),
            media_type="image/jpeg",
        )


def closestNumber(n, m):
    q = int(n / m)
    n1 = m * q
    if (n * m) > 0:
        n2 = m * (q + 1)
    else:
        n2 = m * (q - 1)
    if abs(n - n1) < abs(n - n2):
        return n1
    return n2



import os
import base64
import uuid
import webbrowser

from flask import Flask, request, jsonify, render_template
import requests


tasks = {}

app = Flask(__name__)


@app.route('/')
def index():
    model_name = os.getenv("MODEL_NAME")
    return render_template('index.html', model_name=model_name)


@app.route('/api/generate-image', methods=['POST'])
def generate():
    try:
        data = request.json
        prompt = data.get("prompt")
        if not prompt:
            return jsonify({"status": "failed", "message": "Missing prompt"}), 400

        try:
            params = validate_arguments(data)
        except ValueError as e:
            return jsonify({"status": "failed", "message": f"Invalid params: {data}"}), 400

        print(f"[API] Generating image: {params}")
        task_id = str(uuid.uuid4())
        tasks[task_id] = {"status": "processing"}
        endpoint = os.getenv("MODEL_ENDPOINT")

        # Use a background thread or async task to handle image generation
        generate_image(task_id, params, endpoint)
        return jsonify({"status": "processing", "task_id": task_id}), 202
    except Exception as e:
        return jsonify({"status": "failed", "message": f"Error: {str(e)}"}), 500


@app.route('/api/task-status/<task_id>', methods=['GET'])
def task_status(task_id):
    task = tasks.get(task_id)
    if not task:
        return jsonify({"status": "failed", "message": "Task not found"}), 404
    return jsonify(task)


def generate_image(task_id, params, endpoint):
    try:
        response = requests.get(endpoint, params=params, timeout=300)
        if response.status_code == 200:
            content_type = response.headers.get("Content-Type", "")
            if "image" in content_type:
                image_bytes = response.content
                image_base64 = base64.b64encode(image_bytes).decode()
                tasks[task_id] = {
                    "status": "completed",
                    "image_base64": image_base64
                }
            else:
                tasks[task_id] = {
                    "status": "failed",
                    "message": "Unexpected content type"
                }
        else:
            tasks[task_id] = {
                "status": "failed",
                "message": f"Unexpected status code: {response.status_code}"
            }
    except Exception as e:
        tasks[task_id] = {
            "status": "failed",
            "message": f"Error: {str(e)}"
        }

def validate_and_convert(value, expected_type, default, min_value=None, max_value=None):
    """Convert and validate a value."""
    try:
        if expected_type == int:
            value = int(value)
        elif expected_type == float:
            value = float(value)
        else:
            raise ValueError("Unsupported type for validation")

        if min_value is not None and value < min_value:
            raise ValueError(f"Value must be greater than or equal to {min_value}.")
        if max_value is not None and value > max_value:
            raise ValueError(f"Value must be less than or equal to {max_value}.")

    except (ValueError, TypeError):
        return default

    return value


def validate_arguments(data):
    # Get and validate the values from the data with default values
    width = validate_and_convert(data.get("width", 1024), int, 1024, min_value=1)
    height = validate_and_convert(data.get("height", 1024), int, 1024, min_value=1)
    seed = validate_and_convert(data.get("seed", -1), int, -1)
    guidance_scale = validate_and_convert(data.get("guidance_scale", 3.5), float, 3.5, min_value=0)
    n_steps = validate_and_convert(data.get("n_steps", 20), int, 20, min_value=1)

    # Return validated arguments
    return {
        "prompt": data["prompt"],
        "width": width,
        "height": height,
        "seed": seed,
        "guidance_scale": guidance_scale,
        "n_steps": n_steps
    }


if __name__ == '__main__':
    port = 6868
    webbrowser.open(f"http://localhost:{port}")
    app.run(host='0.0.0.0', port=port)


from setuptools import setup, find_packages

def parse_requirements(filename):
    with open(filename, 'r') as f:
        return [line.strip() for line in f if line.strip() and not line.startswith('#')]

setup(
    name='plumai',
    version='0.1.7',
    py_modules=['plumai'],
    packages=find_packages(include=['plumai', 'plumai.*']),
    install_requires=parse_requirements('requirements.txt'),
    entry_points={
        'console_scripts': [
            'plumai=plumai.plumai:main',
        ],
    },
    author='Tam Nguyen',
    author_email='tamnvhustcc@gmail.com',
    description='A tool for deploying and running AI models to Modal',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://x.com/tamnvvn',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    python_requires='>=3.8',
    package_data={
        '': ['templates/*'],
    },
    include_package_data=True,
)